import abc

# For 2 and 3 compatibility
# https://stackoverflow.com/questions/35673474/
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})


class BaseOfficeFile(ABC):
    def __init__(self):
        pass

    @abc.abstractmethod
    def load_key(self):
        pass

    @abc.abstractmethod
    def decrypt(self):
        pass

    @abc.abstractmethod
    def is_encrypted(self):
        pass



from __future__ import print_function
import logging, sys
import argparse
import getpass

import olefile

from . import __version__
from . import OfficeFile

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def ifWIN32SetBinary(io):
    if sys.platform == 'win32':
        import msvcrt, os
        msvcrt.setmode(io.fileno(), os.O_BINARY)


def is_encrypted(file):
    r'''
    Test if the file is encrypted.

        >>> f = open("tests/inputs/plain.doc", "rb")
        >>> file = OfficeFile(f)
        >>> is_encrypted(file)
        False
    '''
    # TODO: Validate file
    if not olefile.isOleFile(file):
        return False

    file = OfficeFile(file)

    return file.is_encrypted()


parser = argparse.ArgumentParser()
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('-p', '--password', nargs='?', const='', dest='password', help='Password text.')
group.add_argument('-t', '--test', dest='test_encrypted', action='store_true', help='Test if the file is encrypted.')
parser.add_argument('-v', dest='verbose', action='store_true', help='Print verbose information.')
parser.add_argument('infile', nargs='?', type=argparse.FileType('rb'), help='Input file.')
parser.add_argument('outfile', nargs='?', type=argparse.FileType('wb'), help='Output file. If blank, stdout is used.')


def main():
    args = parser.parse_args()

    if args.verbose:
        logger.removeHandler(logging.NullHandler())
        logging.basicConfig(level=logging.DEBUG, format="%(message)s")
        logger.debug("Version: {}".format(__version__))

    if args.test_encrypted:
        if not is_encrypted(args.infile):
            print("{}: not encrypted".format(args.infile.name), file=sys.stderr)
            sys.exit(1)
        else:
            logger.debug("{}: encrypted".format(args.infile.name))
        return

    if not olefile.isOleFile(args.infile):
        raise Exception("Not OLE file")

    file = OfficeFile(args.infile)

    if args.password:
        file.load_key(password=args.password)
    else:
        password = getpass.getpass()
        file.load_key(password=password)

    if args.outfile is None:
        ifWIN32SetBinary(sys.stdout)
        if hasattr(sys.stdout, 'buffer'):  # For Python 2
            args.outfile = sys.stdout.buffer
        else:
            args.outfile = sys.stdout

    file.decrypt(args.outfile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""Compare output of msoffcrypto-tool for a few input files."""

import sys
import unittest
import os
from os.path import dirname, abspath, isfile, join as pjoin
from tempfile import mkstemp
from difflib import SequenceMatcher

try:
    import cryptography
except ImportError:
    cryptography = None

# add base dir to path so we always import local msoffcrypto
TEST_BASE_DIR = dirname(abspath(__file__))
MODULE_BASE_DIR = dirname(TEST_BASE_DIR)
if sys.path[0] != MODULE_BASE_DIR:
    sys.path.insert(0, MODULE_BASE_DIR)
import msoffcrypto


#: encryption password for files tested here
PASSWORD = 'Password1234_'

#: input dir
INPUT_DIR = 'inputs'

#: pairs of input/output files
EXAMPLE_FILES = (
    ('example_password.docx', 'example.docx'),
    ('example_password.xlsx', 'example.xlsx'),
    ('ecma376standard_password.docx', 'ecma376standard_password_plain.docx'),
    ('rc4cryptoapi_password.doc', 'rc4cryptoapi_password_plain.doc'),
    ('rc4cryptoapi_password.xls', 'rc4cryptoapi_password_plain.xls'),
    ('rc4cryptoapi_password.ppt', 'rc4cryptoapi_password_plain.ppt'),
)

#: output dir:
OUTPUT_DIR = 'outputs'



@unittest.skipIf(cryptography is None,
                 'Cryptography module not installed for python{}.{}'
                 .format(sys.version_info.major, sys.version_info.minor))
class KnownOutputCompare(unittest.TestCase):
    """See module doc."""

    def test_known_output(self):
        """See module doc."""
        for in_name, out_name in EXAMPLE_FILES:
            input_path = pjoin(TEST_BASE_DIR, INPUT_DIR, in_name)
            expect_path = pjoin(TEST_BASE_DIR, OUTPUT_DIR, out_name)

            # now run the relevant parts of __main__.main:
            with open(input_path, 'rb') as input_handle:
                file = msoffcrypto.OfficeFile(input_handle)
                if file.format == 'ooxml' and file.type in ['standard', 'agile']:
                    file.load_key(password=PASSWORD, verify_password=True)
                else:
                    file.load_key(password=PASSWORD)

                out_desc = None
                out_path = None
                output = []
                try:
                    # create temp file for output of decryption function
                    out_desc, out_path = mkstemp(prefix='msoffcrypto-test-',
                                                 suffix='.txt', text=True)
                    with os.fdopen(out_desc, 'wb') as out_handle:
                        out_desc = None            # out_handle now owns this

                        # run decryption, capture output
                        print('decrypting {}'.format(in_name))
                        if file.format == 'ooxml' and file.type in ['agile']:
                            file.decrypt(out_handle, verify_integrity=True)
                        else:
                            file.decrypt(out_handle)

                    # read extracted output file into memory
                    with open(expect_path, 'rb') as reader:
                        output = reader.read()
                finally:
                    # ensure we do not leak temp files. Always close & remove
                    if out_desc:
                        os.close(out_desc)
                    if out_path and isfile(out_path):
                        os.unlink(out_path)

            # read output file into memory
            with open(expect_path, 'rb') as reader:
                expect = reader.read()

            # compare:
            print('comparing output to {}'.format(out_name))
            similarity  = SequenceMatcher(None, expect, output).ratio()
            self.assertGreater(similarity, 0.99)


if __name__ == '__main__':
    unittest.main()



import logging, io, shutil, tempfile
from struct import pack, unpack
from collections import namedtuple

import olefile

from . import base
from .common import _parse_encryptionheader, _parse_encryptionverifier
from ..method.rc4 import DocumentRC4
from ..method.rc4_cryptoapi import DocumentRC4CryptoAPI

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


recordNameNum = {
    'Formula': 6,
    'EOF': 10,
    'CalcCount': 12,
    'CalcMode': 13,
    'CalcPrecision': 14,
    'CalcRefMode': 15,
    'CalcDelta': 16,
    'CalcIter': 17,
    'Protect': 18,
    'Password': 19,
    'Header': 20,
    'Footer': 21,
    'ExternSheet': 23,
    'Lbl': 24,
    'WinProtect': 25,
    'VerticalPageBreaks': 26,
    'HorizontalPageBreaks': 27,
    'Note': 28,
    'Selection': 29,
    'Date1904': 34,
    'ExternName': 35,
    'LeftMargin': 38,
    'RightMargin': 39,
    'TopMargin': 40,
    'BottomMargin': 41,
    'PrintRowCol': 42,
    'PrintGrid': 43,
    'FilePass': 47,
    'Font': 49,
    'PrintSize': 51,
    'Continue': 60,
    'Window1': 61,
    'Backup': 64,
    'Pane': 65,
    'CodePage': 66,
    'Pls': 77,
    'DCon': 80,
    'DConRef': 81,
    'DConName': 82,
    'DefColWidth': 85,
    'XCT': 89,
    'CRN': 90,
    'FileSharing': 91,
    'WriteAccess': 92,
    'Obj': 93,
    'Uncalced': 94,
    'CalcSaveRecalc': 95,
    'Template': 96,
    'Intl': 97,
    'ObjProtect': 99,
    'ColInfo': 125,
    'Guts': 128,
    'WsBool': 129,
    'GridSet': 130,
    'HCenter': 131,
    'VCenter': 132,
    'BoundSheet8': 133,
    'WriteProtect': 134,
    'Country': 140,
    'HideObj': 141,
    'Sort': 144,
    'Palette': 146,
    'Sync': 151,
    'LPr': 152,
    'DxGCol': 153,
    'FnGroupName': 154,
    'FilterMode': 155,
    'BuiltInFnGroupCount': 156,
    'AutoFilterInfo': 157,
    'AutoFilter': 158,
    'Scl': 160,
    'Setup': 161,
    'ScenMan': 174,
    'SCENARIO': 175,
    'SxView': 176,
    'Sxvd': 177,
    'SXVI': 178,
    'SxIvd': 180,
    'SXLI': 181,
    'SXPI': 182,
    'DocRoute': 184,
    'RecipName': 185,
    'MulRk': 189,
    'MulBlank': 190,
    'Mms': 193,
    'SXDI': 197,
    'SXDB': 198,
    'SXFDB': 199,
    'SXDBB': 200,
    'SXNum': 201,
    'SxBool': 202,
    'SxErr': 203,
    'SXInt': 204,
    'SXString': 205,
    'SXDtr': 206,
    'SxNil': 207,
    'SXTbl': 208,
    'SXTBRGIITM': 209,
    'SxTbpg': 210,
    'ObProj': 211,
    'SXStreamID': 213,
    'DBCell': 215,
    'SXRng': 216,
    'SxIsxoper': 217,
    'BookBool': 218,
    'DbOrParamQry': 220,
    'ScenarioProtect': 221,
    'OleObjectSize': 222,
    'XF': 224,
    'InterfaceHdr': 225,
    'InterfaceEnd': 226,
    'SXVS': 227,
    'MergeCells': 229,
    'BkHim': 233,
    'MsoDrawingGroup': 235,
    'MsoDrawing': 236,
    'MsoDrawingSelection': 237,
    'PhoneticInfo': 239,
    'SxRule': 240,
    'SXEx': 241,
    'SxFilt': 242,
    'SxDXF': 244,
    'SxItm': 245,
    'SxName': 246,
    'SxSelect': 247,
    'SXPair': 248,
    'SxFmla': 249,
    'SxFormat': 251,
    'SST': 252,
    'LabelSst': 253,
    'ExtSST': 255,
    'SXVDEx': 256,
    'SXFormula': 259,
    'SXDBEx': 290,
    'RRDInsDel': 311,
    'RRDHead': 312,
    'RRDChgCell': 315,
    'RRTabId': 317,
    'RRDRenSheet': 318,
    'RRSort': 319,
    'RRDMove': 320,
    'RRFormat': 330,
    'RRAutoFmt': 331,
    'RRInsertSh': 333,
    'RRDMoveBegin': 334,
    'RRDMoveEnd': 335,
    'RRDInsDelBegin': 336,
    'RRDInsDelEnd': 337,
    'RRDConflict': 338,
    'RRDDefName': 339,
    'RRDRstEtxp': 340,
    'LRng': 351,
    'UsesELFs': 352,
    'DSF': 353,
    'CUsr': 401,
    'CbUsr': 402,
    'UsrInfo': 403,
    'UsrExcl': 404,
    'FileLock': 405,
    'RRDInfo': 406,
    'BCUsrs': 407,
    'UsrChk': 408,
    'UserBView': 425,
    'UserSViewBegin': 426,
    'UserSViewBegin_Chart': 426,
    'UserSViewEnd': 427,
    'RRDUserView': 428,
    'Qsi': 429,
    'SupBook': 430,
    'Prot4Rev': 431,
    'CondFmt': 432,
    'CF': 433,
    'DVal': 434,
    'DConBin': 437,
    'TxO': 438,
    'RefreshAll': 439,
    'HLink': 440,
    'Lel': 441,
    'CodeName': 442,
    'SXFDBType': 443,
    'Prot4RevPass': 444,
    'ObNoMacros': 445,
    'Dv': 446,
    'Excel9File': 448,
    'RecalcId': 449,
    'EntExU2': 450,
    'Dimensions': 512,
    'Blank': 513,
    'Number': 515,
    'Label': 516,
    'BoolErr': 517,
    'String': 519,
    'Row': 520,
    'Index': 523,
    'Array': 545,
    'DefaultRowHeight': 549,
    'Table': 566,
    'Window2': 574,
    'RK': 638,
    'Style': 659,
    'BigName': 1048,
    'Format': 1054,
    'ContinueBigName': 1084,
    'ShrFmla': 1212,
    'HLinkTooltip': 2048,
    'WebPub': 2049,
    'QsiSXTag': 2050,
    'DBQueryExt': 2051,
    'ExtString': 2052,
    'TxtQry': 2053,
    'Qsir': 2054,
    'Qsif': 2055,
    'RRDTQSIF': 2056,
    'BOF': 2057,
    'OleDbConn': 2058,
    'WOpt': 2059,
    'SXViewEx': 2060,
    'SXTH': 2061,
    'SXPIEx': 2062,
    'SXVDTEx': 2063,
    'SXViewEx9': 2064,
    'ContinueFrt': 2066,
    'RealTimeData': 2067,
    'ChartFrtInfo': 2128,
    'FrtWrapper': 2129,
    'StartBlock': 2130,
    'EndBlock': 2131,
    'StartObject': 2132,
    'EndObject': 2133,
    'CatLab': 2134,
    'YMult': 2135,
    'SXViewLink': 2136,
    'PivotChartBits': 2137,
    'FrtFontList': 2138,
    'SheetExt': 2146,
    'BookExt': 2147,
    'SXAddl': 2148,
    'CrErr': 2149,
    'HFPicture': 2150,
    'FeatHdr': 2151,
    'Feat': 2152,
    'DataLabExt': 2154,
    'DataLabExtContents': 2155,
    'CellWatch': 2156,
    'FeatHdr11': 2161,
    'Feature11': 2162,
    'DropDownObjIds': 2164,
    'ContinueFrt11': 2165,
    'DConn': 2166,
    'List12': 2167,
    'Feature12': 2168,
    'CondFmt12': 2169,
    'CF12': 2170,
    'CFEx': 2171,
    'XFCRC': 2172,
    'XFExt': 2173,
    'AutoFilter12': 2174,
    'ContinueFrt12': 2175,
    'MDTInfo': 2180,
    'MDXStr': 2181,
    'MDXTuple': 2182,
    'MDXSet': 2183,
    'MDXProp': 2184,
    'MDXKPI': 2185,
    'MDB': 2186,
    'PLV': 2187,
    'Compat12': 2188,
    'DXF': 2189,
    'TableStyles': 2190,
    'TableStyle': 2191,
    'TableStyleElement': 2192,
    'StyleExt': 2194,
    'NamePublish': 2195,
    'NameCmt': 2196,
    'SortData': 2197,
    'Theme': 2198,
    'GUIDTypeLib': 2199,
    'FnGrp12': 2200,
    'NameFnGrp12': 2201,
    'MTRSettings': 2202,
    'CompressPictures': 2203,
    'HeaderFooter': 2204,
    'CrtLayout12': 2205,
    'CrtMlFrt': 2206,
    'CrtMlFrtContinue': 2207,
    'ForceFullCalculation': 2211,
    'ShapePropsStream': 2212,
    'TextPropsStream': 2213,
    'RichTextStream': 2214,
    'CrtLayout12A': 2215,
    'Units': 4097,
    'Chart': 4098,
    'Series': 4099,
    'DataFormat': 4102,
    'LineFormat': 4103,
    'MarkerFormat': 4105,
    'AreaFormat': 4106,
    'PieFormat': 4107,
    'AttachedLabel': 4108,
    'SeriesText': 4109,
    'ChartFormat': 4116,
    'Legend': 4117,
    'SeriesList': 4118,
    'Bar': 4119,
    'Line': 4120,
    'Pie': 4121,
    'Area': 4122,
    'Scatter': 4123,
    'CrtLine': 4124,
    'Axis': 4125,
    'Tick': 4126,
    'ValueRange': 4127,
    'CatSerRange': 4128,
    'AxisLine': 4129,
    'CrtLink': 4130,
    'DefaultText': 4132,
    'Text': 4133,
    'FontX': 4134,
    'ObjectLink': 4135,
    'Frame': 4146,
    'Begin': 4147,
    'End': 4148,
    'PlotArea': 4149,
    'Chart3d': 4154,
    'PicF': 4156,
    'DropBar': 4157,
    'Radar': 4158,
    'Surf': 4159,
    'RadarArea': 4160,
    'AxisParent': 4161,
    'LegendException': 4163,
    'ShtProps': 4164,
    'SerToCrt': 4165,
    'AxesUsed': 4166,
    'SBaseRef': 4168,
    'SerParent': 4170,
    'SerAuxTrend': 4171,
    'IFmtRecord': 4174,
    'Pos': 4175,
    'AlRuns': 4176,
    'BRAI': 4177,
    'SerAuxErrBar': 4187,
    'ClrtClient': 4188,
    'SerFmt': 4189,
    'Chart3DBarShape': 4191,
    'Fbi': 4192,
    'BopPop': 4193,
    'AxcExt': 4194,
    'Dat': 4195,
    'PlotGrowth': 4196,
    'SIIndex': 4197,
    'GelFrame': 4198,
    'BopPopCustom': 4199,
    'Fbi2': 4200,
}


def _parse_header_RC4(encryptionInfo):
    # RC4: https://msdn.microsoft.com/en-us/library/dd908560(v=office.12).aspx
    salt = encryptionInfo.read(16)
    encryptedVerifier = encryptionInfo.read(16)
    encryptedVerifierHash = encryptionInfo.read(16)
    info = {
        'salt': salt,
        'encryptedVerifier': encryptedVerifier,
        'encryptedVerifierHash': encryptedVerifierHash,
    }
    return info


def _parse_header_RC4CryptoAPI(encryptionInfo):
    flags = encryptionInfo.read(4)
    headerSize, = unpack("<I", encryptionInfo.read(4))
    logger.debug(headerSize)
    blob = io.BytesIO(encryptionInfo.read(headerSize))
    header = _parse_encryptionheader(blob)
    logger.debug(header)
    blob = io.BytesIO(encryptionInfo.read())
    verifier = _parse_encryptionverifier(blob, "RC4")  # TODO: Fix (cf. ooxml.py)
    logger.debug(verifier)
    info = {
        'salt': verifier['salt'],
        'keySize': header['keySize'],
        'encryptedVerifier': verifier['encryptedVerifier'],
        'encryptedVerifierHash': verifier['encryptedVerifierHash'],
    }
    return info


class _BIFFStream:
    def __init__(self, data):
        self.data = data

    def has_record(self, target):
        pos = self.data.tell()
        while True:
            h = self.data.read(4)
            if not h:
                self.data.seek(pos)
                return False
            num, size = unpack("<HH", h)
            if num == target:
                self.data.seek(pos)
                return True
            else:
                self.data.read(size)

    def skip_to(self, target):
        while True:
            h = self.data.read(4)
            if not h:
                raise Exception("Record not found")
            num, size = unpack("<HH", h)
            if num == target:
                return num, size
            else:
                self.data.read(size)

    def iter_record(self):
        while True:
            h = self.data.read(4)
            if not h:
                break
            num, size = unpack("<HH", h)
            record = io.BytesIO(self.data.read(size))
            yield num, size, record


class Xls97File(base.BaseOfficeFile):
    def __init__(self, file):
        self.file = file
        ole = olefile.OleFileIO(file)  # do not close this, would close file
        self.ole = ole
        self.format = "xls97"
        self.keyTypes = ['password']
        self.key = None
        self.salt = None

        workbook = ole.openstream('Workbook')    # closed in destructor

        Data = namedtuple('Data', ['workbook'])
        self.data = Data(
            workbook=workbook,
        )

    def __del__(self):
        """Destructor, closes opened stream."""
        if hasattr(self, 'data') and self.data and self.data.workbook:
            self.data.workbook.close()

    def load_key(self, password=None):
        self.data.workbook.seek(0)
        workbook = _BIFFStream(self.data.workbook)

        # workbook stream consists of records, each of which begins with its ID number.
        # Record IDs (in decimal) are listed here: https://msdn.microsoft.com/en-us/library/dd945945(v=office.12).aspx
        # workbook stream's structure is WORKBOOK = BOF WORKBOOKCONTENT and so forth
        # as in https://msdn.microsoft.com/en-us/library/dd952177(v=office.12).aspx
        # A record begins with its length (in bytes).

        num, = unpack("<H", workbook.data.read(2))

        assert num == 2057  # BOF

        size, = unpack("<H", workbook.data.read(2))
        workbook.data.read(size)  # Skip BOF

        num, size = workbook.skip_to(recordNameNum['FilePass'])  # Skip to FilePass; TODO: Raise exception if not encrypted

        # FilePass: https://msdn.microsoft.com/en-us/library/dd952596(v=office.12).aspx
        # If this record exists, the workbook MUST be encrypted.
        wEncryptionType, = unpack("<H", workbook.data.read(2))

        if wEncryptionType == 0x0000:  # XOR obfuscation
            raise Exception("Unsupported encryption method")
        elif wEncryptionType == 0x0001:  # RC4
            pass

        encryptionInfo = io.BytesIO(workbook.data.read(size-2))

        encryptionVersionInfo = encryptionInfo.read(4)
        vMajor, vMinor = unpack("<HH", encryptionVersionInfo)
        logger.debug("Version: {} {}".format(vMajor, vMinor))
        if vMajor == 0x0001 and vMinor == 0x0001:  # RC4
            info = _parse_header_RC4(encryptionInfo)
            if DocumentRC4.verifypw(password, info['salt'], info['encryptedVerifier'], info['encryptedVerifierHash']):
                self.type = 'rc4'
                self.key = password
                self.salt = info['salt']
            else:
                raise Exception("Failed to verify password")
        elif vMajor in [0x0002, 0x0003, 0x0004] and vMinor == 0x0002:  # RC4 CryptoAPI
            info = _parse_header_RC4CryptoAPI(encryptionInfo)
            if DocumentRC4CryptoAPI.verifypw(password, info['salt'], info['keySize'],
                                             info['encryptedVerifier'], info['encryptedVerifierHash']):
                self.type = 'rc4_cryptoapi'
                self.key = password
                self.salt = info['salt']
                self.keySize = info['keySize']
            else:
                raise Exception("Failed to verify password")
        else:
            raise Exception("Unsupported encryption method")

    def decrypt(self, ofile):
        # fd, _ofile_path = tempfile.mkstemp()

        # shutil.copyfile(os.path.realpath(self.file.name), _ofile_path)
        # outole = olefile.OleFileIO(_ofile_path, write_mode=True)

        # List of encrypted parts: https://msdn.microsoft.com/en-us/library/dd905723(v=office.12).aspx

        # Workbook stream
        self.data.workbook.seek(0)
        workbook = _BIFFStream(self.data.workbook)

        plain_buf = []
        encrypted_buf = io.BytesIO()

        for i, (num, size, record) in enumerate(workbook.iter_record()):
            # Remove encryption, pad by zero to preserve stream size
            if num == recordNameNum['FilePass']:
                plain_buf += [0, 0] + list(pack("<H", size)) + [0] * size
                encrypted_buf.write(b"\x00"*(4+size))
            # The following records MUST NOT be obfuscated or encrypted: BOF (section 2.4.21),
            # FilePass (section 2.4.117), UsrExcl (section 2.4.339), FileLock (section 2.4.116),
            # InterfaceHdr (section 2.4.146), RRDInfo (section 2.4.227), and RRDHead (section 2.4.226).
            elif num in [recordNameNum['BOF'], recordNameNum['FilePass'], recordNameNum['UsrExcl'],
                         recordNameNum['FileLock'], recordNameNum['InterfaceHdr'], recordNameNum['RRDInfo'],
                         recordNameNum['RRDHead']]:
                header = pack("<HH", num, size)
                plain_buf += list(header) + list(record.read())
                encrypted_buf.write(b"\x00"*(4+size))
            # The lbPlyPos field of the BoundSheet8 record (section 2.4.28) MUST NOT be encrypted.
            elif num == recordNameNum['BoundSheet8']:
                header = pack("<HH", num, size)
                plain_buf += list(header) + list(record.read(4)) + [-1] * (size-4)  # Preserve lbPlyPos
                encrypted_buf.write(b"\x00"*4 + b"\x00"*4 + record.read())
            else:
                header = pack("<HH", num, size)
                plain_buf += list(header) + [-1] * size
                encrypted_buf.write(b"\x00"*4 + record.read())

        encrypted_buf.seek(0)

        if self.type == "rc4":
            dec = DocumentRC4.decrypt(self.key, self.salt, encrypted_buf, blocksize=1024)
        elif self.type == "rc4_cryptoapi":
            dec = DocumentRC4CryptoAPI.decrypt(self.key, self.salt, self.keySize, encrypted_buf, blocksize=1024)

        for c in plain_buf:
            if c == -1:
                dec.seek(1, 1)
            else:
                dec.write(bytearray([c]))

        dec.seek(0)

        # f = open('Workbook', 'wb')
        # f.write(dec.read())
        # dec.seek(0)

        workbook_dec = dec

        with tempfile.TemporaryFile() as _ofile:
            self.file.seek(0)
            shutil.copyfileobj(self.file, _ofile)
            outole = olefile.OleFileIO(_ofile, write_mode=True)

            outole.write_stream('Workbook', workbook_dec.read())

            # _ofile = open(_ofile_path, 'rb')

            _ofile.seek(0)

            shutil.copyfileobj(_ofile, ofile)

        return

    def is_encrypted(self):
        r'''
        Test if the file is encrypted.

            >>> f = open("tests/inputs/plain.xls", "rb")
            >>> file = Xls97File(f)
            >>> file.is_encrypted()
            False
            >>> f = open("tests/inputs/rc4cryptoapi_password.xls", "rb")
            >>> file = Xls97File(f)
            >>> file.is_encrypted()
            True
        '''
        # Utilising the method above, check for encryption type.
        self.data.workbook.seek(0)
        workbook = _BIFFStream(self.data.workbook)

        num, = unpack("<H", workbook.data.read(2))
        assert num == 2057

        size, = unpack("<H", workbook.data.read(2))
        workbook.data.read(size)

        if not workbook.has_record(recordNameNum['FilePass']):
            return False

        num, size = workbook.skip_to(recordNameNum['FilePass'])
        wEncryptionType, = unpack("<H", workbook.data.read(2))

        if wEncryptionType == 0x0001:  # RC4
            return True
        elif wEncryptionType == 0x0000:  # XOR obfuscation
            # If not compatible no point stating that
            raise Exception("Unsupported encryption method")
        else:
            return False



import logging, io, shutil, tempfile
from struct import pack, unpack
from collections import namedtuple

import olefile

from . import base
from .common import _parse_encryptionheader, _parse_encryptionverifier
from ..method.rc4_cryptoapi import DocumentRC4CryptoAPI

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


RecordHeader = namedtuple('RecordHeader', [
    'recVer',
    'recInstance',
    'recType',
    'recLen',
])


def _parseRecordHeader(blob):
    # RecordHeader: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/df201194-0cd0-4dfb-bf10-eea353d8eabc
    getBitSlice = lambda bits, i, w: (bits & (2 ** w - 1 << i)) >> i

    blob.seek(0)

    buf, = unpack("<H", blob.read(2))
    recVer = getBitSlice(buf, 0, 4)
    recInstance = getBitSlice(buf, 4, 12)

    recType, = unpack("<H", blob.read(2))
    recLen, = unpack("<I", blob.read(4))

    rh = RecordHeader(
        recVer=recVer,
        recInstance=recInstance,
        recType=recType,
        recLen=recLen,
    )

    return rh


def _packRecordHeader(rh):
    setBitSlice = lambda bits, i, w, v: (bits & ~((2**w - 1) << i)) | ((v & (2**w - 1)) << i)

    blob = io.BytesIO()

    _buf = 0xffff
    _buf = setBitSlice(_buf, 0, 4, rh.recVer)
    _buf = setBitSlice(_buf, 4, 12, rh.recInstance)
    buf = pack("<H", _buf)
    blob.write(buf)

    buf = pack("<H", rh.recType)
    blob.write(buf)

    buf = pack("<I", rh.recLen)
    blob.write(buf)

    blob.seek(0)

    return blob


CurrentUserAtom = namedtuple('CurrentUserAtom', [
    'rh',
    'size',
    'headerToken',
    'offsetToCurrentEdit',
    'lenUserName',
    'docFileVersion',
    'majorVersion',
    'minorVersion',
    'unused',
    'ansiUserName',
    'relVersion',
    'unicodeUserName',
])


def _parseCurrentUserAtom(blob):
    # CurrentUserAtom: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/940d5700-e4d7-4fc0-ab48-fed5dbc48bc1

    # rh (8 bytes): A RecordHeader structure...
    buf = io.BytesIO(blob.read(8))
    rh = _parseRecordHeader(buf)
    # logger.debug(rh)

    # ...Sub-fields are further specified in the following table.
    assert rh.recVer == 0x0
    assert rh.recInstance == 0x000
    assert rh.recType == 0x0ff6

    size, = unpack("<I", blob.read(4))
    # logger.debug(hex(size))

    # size (4 bytes): ...It MUST be 0x00000014.
    assert size == 0x00000014

    # headerToken (4 bytes): An unsigned integer that specifies
    # a token used to identify whether the file is encrypted.
    headerToken, = unpack("<I", blob.read(4))

    # TODO: Check headerToken value

    offsetToCurrentEdit, = unpack("<I", blob.read(4))

    lenUserName, = unpack("<H", blob.read(2))
    docFileVersion, = unpack("<H", blob.read(2))
    majorVersion, minorVersion, = unpack("<BB", blob.read(2))
    unused = blob.read(2)
    ansiUserName = blob.read(lenUserName)
    relVersion, = unpack("<I", blob.read(4))
    unicodeUserName = blob.read(2*lenUserName)

    return CurrentUserAtom(
        rh=rh,
        size=size,
        headerToken=headerToken,
        offsetToCurrentEdit=offsetToCurrentEdit,
        lenUserName=lenUserName,
        docFileVersion=docFileVersion,
        majorVersion=majorVersion,
        minorVersion=minorVersion,
        unused=unused,
        ansiUserName=ansiUserName,
        relVersion=relVersion,
        unicodeUserName=unicodeUserName,
    )


def _packCurrentUserAtom(currentuseratom):
    blob = io.BytesIO()

    buf = _packRecordHeader(currentuseratom.rh).read()
    blob.write(buf)
    buf = pack("<I", currentuseratom.size)
    blob.write(buf)
    buf = pack("<I", currentuseratom.headerToken)
    blob.write(buf)
    buf = pack("<I", currentuseratom.offsetToCurrentEdit)
    blob.write(buf)
    buf = pack("<H", currentuseratom.lenUserName)
    blob.write(buf)
    buf = pack("<H", currentuseratom.docFileVersion)
    blob.write(buf)
    buf = pack("<BB", currentuseratom.majorVersion, currentuseratom.minorVersion)
    blob.write(buf)
    buf = currentuseratom.unused
    blob.write(buf)
    buf = currentuseratom.ansiUserName
    blob.write(buf)
    buf = pack("<I", currentuseratom.relVersion)
    blob.write(buf)
    buf = currentuseratom.unicodeUserName
    blob.write(buf)

    blob.seek(0)

    return blob


CurrentUser = namedtuple('CurrentUser', ['currentuseratom'])


def _parseCurrentUser(blob):
    # Current User Stream: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/76cfa657-07a6-464b-81ab-4c017c611f64
    currentuser = CurrentUser(
        currentuseratom=_parseCurrentUserAtom(blob)
    )
    return currentuser


def _packCurrentUser(currentuser):
    blob = io.BytesIO()

    buf = _packCurrentUserAtom(currentuser.currentuseratom).read()
    blob.write(buf)

    blob.seek(0)

    return blob


UserEditAtom = namedtuple('UserEditAtom', [
    'rh',
    'lastSlideIdRef',
    'version',
    'minorVersion',
    'majorVersion',
    'offsetLastEdit',
    'offsetPersistDirectory',
    'docPersistIdRef',
    'persistIdSeed',
    'lastView',
    'unused',
    'encryptSessionPersistIdRef',
])


def _parseUserEditAtom(blob):
    # UserEditAtom: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/3ffb3fab-95de-4873-98aa-d508fbbac981

    # rh (8 bytes): A RecordHeader structure...
    buf = io.BytesIO(blob.read(8))
    rh = _parseRecordHeader(buf)
    # logger.debug(rh)

    # ...Sub-fields are further specified in the following table.
    assert rh.recVer == 0x0
    assert rh.recInstance == 0x000
    assert rh.recType == 0x0ff5
    assert rh.recLen == 0x0000001c or rh.recLen == 0x00000020  # 0x0000001c + len(encryptSessionPersistIdRef)

    lastSlideIdRef, = unpack("<I", blob.read(4))
    version, = unpack("<H", blob.read(2))
    minorVersion, majorVersion, = unpack("<BB", blob.read(2))
    # majorVersion, minorVersion, = unpack("<BB", blob.read(2))

    offsetLastEdit, = unpack("<I", blob.read(4))
    offsetPersistDirectory, = unpack("<I", blob.read(4))
    docPersistIdRef, = unpack("<I", blob.read(4))

    persistIdSeed, = unpack("<I", blob.read(4))
    lastView, = unpack("<H", blob.read(2))
    unused = blob.read(2)

    # encryptSessionPersistIdRef (4 bytes): An optional PersistIdRef
    # that specifies the value to look up in the persist object directory
    # to find the offset of the CryptSession10Container record (section 2.3.7).
    buf = blob.read(4)
    if len(buf) == 4:
        encryptSessionPersistIdRef, = unpack("<I", buf)
    else:
        encryptSessionPersistIdRef = None

    return UserEditAtom(
        rh=rh,
        lastSlideIdRef=lastSlideIdRef,
        version=version,
        minorVersion=minorVersion,
        majorVersion=majorVersion,
        offsetLastEdit=offsetLastEdit,
        offsetPersistDirectory=offsetPersistDirectory,
        docPersistIdRef=docPersistIdRef,
        persistIdSeed=persistIdSeed,
        lastView=lastView,
        unused=unused,
        encryptSessionPersistIdRef=encryptSessionPersistIdRef,
    )


def _packUserEditAtom(usereditatom):
    blob = io.BytesIO()

    buf = _packRecordHeader(usereditatom.rh).read()
    blob.write(buf)
    buf = pack("<I", usereditatom.lastSlideIdRef)
    blob.write(buf)
    buf = pack("<H", usereditatom.version)
    blob.write(buf)
    buf = pack("<BB", usereditatom.minorVersion, usereditatom.majorVersion)
    blob.write(buf)
    buf = pack("<I", usereditatom.offsetLastEdit)
    blob.write(buf)
    buf = pack("<I", usereditatom.offsetPersistDirectory)
    blob.write(buf)
    buf = pack("<I", usereditatom.docPersistIdRef)
    blob.write(buf)
    buf = pack("<I", usereditatom.persistIdSeed)
    blob.write(buf)
    buf = pack("<H", usereditatom.lastView)
    blob.write(buf)
    buf = usereditatom.unused
    blob.write(buf)
    # Optional value
    if usereditatom.encryptSessionPersistIdRef is not None:
        buf = pack("<I", usereditatom.encryptSessionPersistIdRef)
        blob.write(buf)

    blob.seek(0)

    return blob


PersistDirectoryEntry = namedtuple('PersistDirectoryEntry', [
    'persistId',
    'cPersist',
    'rgPersistOffset',
])


def _parsePersistDirectoryEntry(blob):
    # PersistDirectoryEntry: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/6214b5a6-7ca2-4a86-8a0e-5fd3d3eff1c9
    getBitSlice = lambda bits, i, w: (bits & (2 ** w - 1 << i)) >> i

    buf, = unpack("<I", blob.read(4))
    persistId = getBitSlice(buf, 0, 20)
    cPersist = getBitSlice(buf, 20, 12)

    # cf. PersistOffsetEntry: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/a056484a-2132-4e1e-aa54-6e387f9695cf
    size_rgPersistOffset = 4 * cPersist
    _rgPersistOffset = blob.read(size_rgPersistOffset)
    _rgPersistOffset = io.BytesIO(_rgPersistOffset)
    rgPersistOffset = []
    pos = 0
    while pos < size_rgPersistOffset:
        persistoffsetentry, = unpack("<I", _rgPersistOffset.read(4))
        rgPersistOffset.append(persistoffsetentry)
        pos += 4

    return PersistDirectoryEntry(
        persistId=persistId,
        cPersist=cPersist,
        rgPersistOffset=rgPersistOffset,
    )


def _packPersistDirectoryEntry(directoryentry):
    setBitSlice = lambda bits, i, w, v: (bits & ~((2**w - 1) << i)) | ((v & (2**w - 1)) << i)

    blob = io.BytesIO()

    _buf = 0xffffffff
    _buf = setBitSlice(_buf, 0, 20, directoryentry.persistId)
    _buf = setBitSlice(_buf, 20, 12, directoryentry.cPersist)
    buf = pack("<I", _buf)
    blob.write(buf)

    for v in directoryentry.rgPersistOffset:
        buf = pack("<I", v)
        blob.write(buf)

    blob.seek(0)

    return blob


PersistDirectoryAtom = namedtuple('PersistDirectoryAtom', [
    'rh',
    'rgPersistDirEntry',
])


def _parsePersistDirectoryAtom(blob):
    # PersistDirectoryAtom: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/d10a093d-860f-409c-b065-aeb24b830505

    # rh (8 bytes): A RecordHeader structure...
    buf = io.BytesIO(blob.read(8))
    rh = _parseRecordHeader(buf)
    # logger.debug(rh)

    # ...Sub-fields are further specified in the following table.
    assert rh.recVer == 0x0
    assert rh.recInstance == 0x000
    assert rh.recType == 0x1772

    _rgPersistDirEntry = blob.read(rh.recLen)
    _rgPersistDirEntry = io.BytesIO(_rgPersistDirEntry)
    rgPersistDirEntry = []
    pos = 0
    while pos < rh.recLen:
        persistdirectoryentry = _parsePersistDirectoryEntry(_rgPersistDirEntry)
        size_persistdirectoryentry = 4 + 4 * len(persistdirectoryentry.rgPersistOffset)
        # logger.debug((persistdirectoryentry, size_persistdirectoryentry))
        rgPersistDirEntry.append(persistdirectoryentry)
        pos += size_persistdirectoryentry

    return PersistDirectoryAtom(
        rh=rh,
        rgPersistDirEntry=rgPersistDirEntry,
    )


def _packPersistDirectoryAtom(directoryatom):
    blob = io.BytesIO()

    buf = _packRecordHeader(directoryatom.rh).read()
    blob.write(buf)

    for v in directoryatom.rgPersistDirEntry:
        buf = _packPersistDirectoryEntry(v)
        blob.write(buf.read())

    blob.seek(0)

    return blob


def _parseCryptSession10Container(blob):
    # CryptSession10Container: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/b0963334-4408-4621-879a-ef9c54551fd8

    CryptSession10Container = namedtuple('CryptSession10Container', [
        'rh',
        'data',
    ])

    # rh (8 bytes): A RecordHeader structure...
    buf = io.BytesIO(blob.read(8))
    rh = _parseRecordHeader(buf)
    # logger.debug(rh)

    # ...Sub-fields are further specified in the following table.
    assert rh.recVer == 0xf
    # The specified value fails
    # assert rh.recInstance == 0x000
    assert rh.recType == 0x2f14

    data = blob.read(rh.recLen)

    return CryptSession10Container(
        rh=rh,
        data=data,
    )


def construct_persistobjectdirectory(data):
    # PowerPoint Document Stream: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/1fc22d56-28f9-4818-bd45-67c2bf721ccf

    # 1. Read the CurrentUserAtom record (section 2.3.2) from the Current User Stream (section 2.1.1). ...
    data.currentuser.seek(0)
    currentuser = _parseCurrentUser(data.currentuser)
    # logger.debug(currentuser)

    # 2. Seek, in the PowerPoint Document Stream, to the offset specified by the offsetToCurrentEdit field of
    # the CurrentUserAtom record identified in step 1.
    data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)

    persistdirectoryatom_stack = []

    # The stream MUST contain exactly one UserEditAtom record.
    # https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/b0963334-4408-4621-879a-ef9c54551fd8
    for i in range(1):
        # 3. Read the UserEditAtom record at the current offset. ...
        usereditatom = _parseUserEditAtom(data.powerpointdocument)
        # logger.debug(usereditatom)

        # 4. Seek to the offset specified by the offsetPersistDirectory field of the UserEditAtom record identified in step 3.
        data.powerpointdocument.seek(usereditatom.offsetPersistDirectory)

        # 5. Read the PersistDirectoryAtom record at the current offset. ...
        persistdirectoryatom = _parsePersistDirectoryAtom(data.powerpointdocument)
        # logger.debug(persistdirectoryatom)
        persistdirectoryatom_stack.append(persistdirectoryatom)

        # 6. Seek to the offset specified by the offsetLastEdit field in the UserEditAtom record identified in step 3.
        # 7. Repeat steps 3 through 6 until offsetLastEdit is 0x00000000.
        if usereditatom.offsetLastEdit == 0x00000000:
            break
        else:
            data.powerpointdocument.seek(usereditatom.offsetLastEdit)

    # 8. Construct the complete persist object directory for this file as follows:
    persistobjectdirectory = {}

    # 8a. For each PersistDirectoryAtom record previously identified in step 5,
    # add the persist object identifier and persist object stream offset pairs to
    # the persist object directory starting with the PersistDirectoryAtom record
    # last identified, that is, the one closest to the beginning of the stream.
    # 8b. Continue adding these pairs to the persist object directory for each PersistDirectoryAtom record
    # in the reverse order that they were identified in step 5; that is, the pairs from the PersistDirectoryAtom record
    # closest to the end of the stream are added last.
    # 8c. When adding a new pair to the persist object directory, if the persist object identifier
    # already exists in the persist object directory, the persist object stream offset from
    # the new pair replaces the existing persist object stream offset for that persist object identifier.
    while len(persistdirectoryatom_stack) > 0:
        persistdirectoryatom = persistdirectoryatom_stack.pop()
        for entry in persistdirectoryatom.rgPersistDirEntry:
            # logger.debug("persistId: %d" % entry.persistId)
            for i, offset in enumerate(entry.rgPersistOffset):
                persistobjectdirectory[entry.persistId + i] = offset

    return persistobjectdirectory


def _parse_header_RC4CryptoAPI(encryptionInfo):
    flags = encryptionInfo.read(4)
    headerSize, = unpack("<I", encryptionInfo.read(4))
    logger.debug(headerSize)
    blob = io.BytesIO(encryptionInfo.read(headerSize))
    header = _parse_encryptionheader(blob)
    logger.debug(header)
    blob = io.BytesIO(encryptionInfo.read())
    verifier = _parse_encryptionverifier(blob, "RC4")  # TODO: Fix (cf. ooxml.py)
    logger.debug(verifier)
    info = {
        'salt': verifier['salt'],
        'keySize': header['keySize'],
        'encryptedVerifier': verifier['encryptedVerifier'],
        'encryptedVerifierHash': verifier['encryptedVerifierHash'],
    }
    return info


class Ppt97File(base.BaseOfficeFile):
    def __init__(self, file):
        self.file = file
        ole = olefile.OleFileIO(file)  # do not close this, would close file
        self.ole = ole
        self.format = "ppt97"
        self.keyTypes = ['password']
        self.key = None
        self.salt = None

        # streams closed in destructor:
        currentuser = ole.openstream('Current User')
        powerpointdocument = ole.openstream('PowerPoint Document')

        Data = namedtuple('Data', ['currentuser', 'powerpointdocument'])
        self.data = Data(
            currentuser=currentuser,
            powerpointdocument=powerpointdocument,
        )

    def __del__(self):
        """Destructor, closes opened streams."""
        if hasattr(self, 'data') and self.data:
            if self.data.currentuser:
                self.data.currentuser.close()
            if self.data.powerpointdocument:
                self.data.powerpointdocument.close()

    def load_key(self, password=None):
        persistobjectdirectory = construct_persistobjectdirectory(self.data)
        logger.debug("[*] persistobjectdirectory: {}".format(persistobjectdirectory))

        self.data.currentuser.seek(0)
        currentuser = _parseCurrentUser(self.data.currentuser)
        logger.debug("[*] currentuser: {}".format(currentuser))

        self.data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)
        usereditatom = _parseUserEditAtom(self.data.powerpointdocument)
        logger.debug("[*] usereditatom: {}".format(usereditatom))

        # cf. Part 2 in https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/1fc22d56-28f9-4818-bd45-67c2bf721ccf
        cryptsession10container_offset = persistobjectdirectory[usereditatom.encryptSessionPersistIdRef]
        logger.debug("[*] cryptsession10container_offset: {}".format(cryptsession10container_offset))

        self.data.powerpointdocument.seek(cryptsession10container_offset)
        cryptsession10container = _parseCryptSession10Container(self.data.powerpointdocument)
        logger.debug("[*] cryptsession10container: {}".format(cryptsession10container))

        encryptionInfo = io.BytesIO(cryptsession10container.data)

        encryptionVersionInfo = encryptionInfo.read(4)
        vMajor, vMinor = unpack("<HH", encryptionVersionInfo)
        logger.debug("[*] encryption version: {} {}".format(vMajor, vMinor))

        assert vMajor in [0x0002, 0x0003, 0x0004] and vMinor == 0x0002  # RC4 CryptoAPI

        info = _parse_header_RC4CryptoAPI(encryptionInfo)
        if DocumentRC4CryptoAPI.verifypw(password, info['salt'], info['keySize'],
                                         info['encryptedVerifier'], info['encryptedVerifierHash']):
            self.type = 'rc4_cryptoapi'
            self.key = password
            self.salt = info['salt']
            self.keySize = info['keySize']
        else:
            raise Exception("Failed to verify password")

    def decrypt(self, ofile):
        # Current User Stream
        self.data.currentuser.seek(0)
        currentuser = _parseCurrentUser(self.data.currentuser)
        # logger.debug(currentuser)

        cuatom = currentuser.currentuseratom

        currentuser_new = CurrentUser(
            currentuseratom=CurrentUserAtom(
                rh=cuatom.rh,
                size=cuatom.size,
                # https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/940d5700-e4d7-4fc0-ab48-fed5dbc48bc1
                # 0xE391C05F: The file SHOULD NOT<6> be an encrypted document.
                headerToken=0xe391c05f,
                offsetToCurrentEdit=cuatom.offsetToCurrentEdit,
                lenUserName=cuatom.lenUserName,
                docFileVersion=cuatom.docFileVersion,
                majorVersion=cuatom.majorVersion,
                minorVersion=cuatom.minorVersion,
                unused=cuatom.unused,
                ansiUserName=cuatom.ansiUserName,
                relVersion=cuatom.relVersion,
                unicodeUserName=cuatom.unicodeUserName,
            )
        )

        buf = _packCurrentUser(currentuser_new)
        buf.seek(0)
        currentuser_buf = buf

        # List of encrypted parts: https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-ppt/b0963334-4408-4621-879a-ef9c54551fd8

        # PowerPoint Document Stream

        self.data.powerpointdocument.seek(0)
        powerpointdocument_size = len(self.data.powerpointdocument.read())
        logger.debug("[*] powerpointdocument_size: {}".format(powerpointdocument_size))

        self.data.powerpointdocument.seek(0)
        dec_bytearray = bytearray(self.data.powerpointdocument.read())

        # UserEditAtom
        self.data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)
        # currentuseratom_raw = self.data.powerpointdocument.read(40)

        self.data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)
        usereditatom = _parseUserEditAtom(self.data.powerpointdocument)
        # logger.debug(usereditatom)
        # logger.debug(["offsetToCurrentEdit", currentuser.currentuseratom.offsetToCurrentEdit])

        rh_new = RecordHeader(
            recVer=usereditatom.rh.recVer,
            recInstance=usereditatom.rh.recInstance,
            recType=usereditatom.rh.recType,
            recLen=usereditatom.rh.recLen - 4,  # Omit encryptSessionPersistIdRef field
        )

        # logger.debug([_packRecordHeader(usereditatom.rh).read(), _packRecordHeader(rh_new).read()])

        usereditatom_new = UserEditAtom(
            rh=rh_new,
            lastSlideIdRef=usereditatom.lastSlideIdRef,
            version=usereditatom.version,
            minorVersion=usereditatom.minorVersion,
            majorVersion=usereditatom.majorVersion,
            offsetLastEdit=usereditatom.offsetLastEdit,
            offsetPersistDirectory=usereditatom.offsetPersistDirectory,
            docPersistIdRef=usereditatom.docPersistIdRef,
            persistIdSeed=usereditatom.persistIdSeed,
            lastView=usereditatom.lastView,
            unused=usereditatom.unused,
            encryptSessionPersistIdRef=0x00000000,  # Clear
        )

        # logger.debug(currentuseratom_raw)
        # logger.debug(_packUserEditAtom(usereditatom).read())
        # logger.debug(_packUserEditAtom(usereditatom_new).read())

        buf = _packUserEditAtom(usereditatom_new)
        buf.seek(0)
        buf_bytes = bytearray(buf.read())
        offset = currentuser.currentuseratom.offsetToCurrentEdit
        dec_bytearray[offset:offset+len(buf_bytes)] = buf_bytes

        # PersistDirectoryAtom
        self.data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)
        usereditatom = _parseUserEditAtom(self.data.powerpointdocument)
        # logger.debug(usereditatom)

        self.data.powerpointdocument.seek(usereditatom.offsetPersistDirectory)
        persistdirectoryatom = _parsePersistDirectoryAtom(self.data.powerpointdocument)
        # logger.debug(persistdirectoryatom)

        persistdirectoryatom_new = PersistDirectoryAtom(
            rh=persistdirectoryatom.rh,
            rgPersistDirEntry=[
                PersistDirectoryEntry(
                    persistId=persistdirectoryatom.rgPersistDirEntry[0].persistId,
                    # Omit CryptSession10Container
                    cPersist=persistdirectoryatom.rgPersistDirEntry[0].cPersist-1,
                    rgPersistOffset=persistdirectoryatom.rgPersistDirEntry[0].rgPersistOffset,
                ),
            ],
        )

        self.data.powerpointdocument.seek(usereditatom.offsetPersistDirectory)
        buf = _packPersistDirectoryAtom(persistdirectoryatom_new)
        buf_bytes = bytearray(buf.read())
        offset = usereditatom.offsetPersistDirectory
        dec_bytearray[offset:offset+len(buf_bytes)] = buf_bytes

        # Persist Objects
        self.data.powerpointdocument.seek(0)
        persistobjectdirectory = construct_persistobjectdirectory(self.data)

        directory_items = list(persistobjectdirectory.items())

        for i, (persistId, offset) in enumerate(directory_items):
            self.data.powerpointdocument.seek(offset)
            buf = self.data.powerpointdocument.read(8)
            rh = _parseRecordHeader(io.BytesIO(buf))
            logger.debug("[*] rh: {}".format(rh))

            # CryptSession10Container
            if rh.recType == 0x2f14:
                logger.debug("[*] CryptSession10Container found")
                # Remove encryption, pad by zero to preserve stream size
                dec_bytearray[offset:offset+(8+rh.recLen)] = b"\x00" * (8+rh.recLen)
                continue

            # The UserEditAtom record (section 2.3.3) and the PersistDirectoryAtom record (section 2.3.4) MUST NOT be encrypted.
            if rh.recType in [0x0ff5, 0x1772]:
                logger.debug("[*] UserEditAtom/PersistDirectoryAtom found")
                continue

            # TODO: Fix here
            recLen = directory_items[i+1][1] - offset - 8
            logger.debug("[*] recLen: {}".format(recLen))

            self.data.powerpointdocument.seek(offset)
            enc_buf = io.BytesIO(self.data.powerpointdocument.read(8+recLen))
            blocksize = self.keySize * ((8 + recLen) // self.keySize + 1)  # Undocumented
            dec = DocumentRC4CryptoAPI.decrypt(self.key, self.salt, self.keySize, enc_buf, blocksize=blocksize, block=persistId)
            dec_bytes = bytearray(dec.read())
            dec_bytearray[offset:offset+len(dec_bytes)] = dec_bytes

        # To BytesIO
        dec_buf = io.BytesIO(dec_bytearray)

        dec_buf.seek(0)
        for i, (persistId, offset) in enumerate(directory_items):
            dec_buf.seek(offset)
            buf = dec_buf.read(8)
            rh = _parseRecordHeader(io.BytesIO(buf))
            logger.debug("[*] rh: {}".format(rh))

        dec_buf.seek(0)
        logger.debug("[*] powerpointdocument_size={}, len(dec_buf.read())={}".format(powerpointdocument_size, len(dec_buf.read())))

        dec_buf.seek(0)
        powerpointdocument_dec_buf = dec_buf

        # TODO: Pictures Stream
        # TODO: Encrypted Summary Info Stream

        with tempfile.TemporaryFile() as _ofile:
            self.file.seek(0)
            shutil.copyfileobj(self.file, _ofile)
            outole = olefile.OleFileIO(_ofile, write_mode=True)

            outole.write_stream('Current User', currentuser_buf.read())
            outole.write_stream('PowerPoint Document', powerpointdocument_dec_buf.read())

            # Finalize
            _ofile.seek(0)
            shutil.copyfileobj(_ofile, ofile)

        return

    def is_encrypted(self):
        r'''
        Test if the file is encrypted.

            >>> f = open("tests/inputs/plain.ppt", "rb")
            >>> file = Ppt97File(f)
            >>> file.is_encrypted()
            False
            >>> f = open("tests/inputs/rc4cryptoapi_password.ppt", "rb")
            >>> file = Ppt97File(f)
            >>> file.is_encrypted()
            True
        '''
        self.data.currentuser.seek(0)
        currentuser = _parseCurrentUser(self.data.currentuser)
        logger.debug("[*] currentuser: {}".format(currentuser))

        self.data.powerpointdocument.seek(currentuser.currentuseratom.offsetToCurrentEdit)
        usereditatom = _parseUserEditAtom(self.data.powerpointdocument)
        logger.debug("[*] usereditatom: {}".format(usereditatom))

        if usereditatom.rh.recLen == 0x00000020:  # Cf. _parseUserEditAtom
            return True
        else:
            return False



"""Check that given file handles are not closed."""


import unittest
from os.path import join, dirname

from msoffcrypto import OfficeFile


#: directory with input
DATA_DIR = join(dirname(__file__), 'inputs')


class FileHandleTest(unittest.TestCase):
    """See module doc."""

    def test_file_handle_open(self):
        """Check that file handles are open after is_encrypted()."""
        for suffix in 'doc', 'ppt', 'xls':
            path = join(DATA_DIR, 'plain.' + suffix)

            with open(path, 'rb') as file_handle:
                ofile = OfficeFile(file_handle)

                # do something with ofile
                self.assertEqual(ofile.is_encrypted(), False)

                # check that file handle is still open
                self.assertFalse(file_handle.closed)

                # destroy OfficeFile, calls destructor
                del ofile

                # check that file handle is still open
                self.assertFalse(file_handle.closed)

            # just for completeness:
            # check that file handle is now closed
            self.assertTrue(file_handle.closed)


# if someone calls this as script, run unittests
if __name__ == '__main__':
    unittest.main()



import logging
import base64, io
from struct import unpack
from xml.dom.minidom import parseString
import zipfile

import olefile

from . import base
from .common import _parse_encryptionheader, _parse_encryptionverifier
from ..method.ecma376_agile import ECMA376Agile
from ..method.ecma376_standard import ECMA376Standard

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _parseinfo_standard(ole):
    headerFlags, = unpack('<I', ole.read(4))
    encryptionHeaderSize, = unpack('<I', ole.read(4))
    block = ole.read(encryptionHeaderSize)
    blob = io.BytesIO(block)
    header = _parse_encryptionheader(blob)
    block = ole.read()
    blob = io.BytesIO(block)
    algIdMap = {
        0x0000660E: 'AES-128',
        0x0000660F: 'AES-192',
        0x00006610: 'AES-256',
    }
    verifier = _parse_encryptionverifier(blob, "AES" if header['algId'] & 0xFF00 == 0x6600 else "RC4")  # TODO: Fix
    info = {
        'header': header,
        'verifier': verifier,
    }
    return info


def _parseinfo_agile(ole):
    ole.seek(8)
    xml = parseString(ole.read())
    keyDataSalt = base64.b64decode(xml.getElementsByTagName('keyData')[0].getAttribute('saltValue'))
    keyDataHashAlgorithm = xml.getElementsByTagName('keyData')[0].getAttribute('hashAlgorithm')
    keyDataBlockSize = int(xml.getElementsByTagName('keyData')[0].getAttribute('blockSize'))
    encryptedHmacKey = base64.b64decode(xml.getElementsByTagName('dataIntegrity')[0].getAttribute('encryptedHmacKey'))
    encryptedHmacValue = base64.b64decode(xml.getElementsByTagName('dataIntegrity')[0].getAttribute('encryptedHmacValue'))
    password_node = xml.getElementsByTagNameNS("http://schemas.microsoft.com/office/2006/keyEncryptor/password", 'encryptedKey')[0]
    spinValue = int(password_node.getAttribute('spinCount'))
    encryptedKeyValue = base64.b64decode(password_node.getAttribute('encryptedKeyValue'))
    encryptedVerifierHashInput = base64.b64decode(password_node.getAttribute('encryptedVerifierHashInput'))
    encryptedVerifierHashValue = base64.b64decode(password_node.getAttribute('encryptedVerifierHashValue'))
    passwordSalt = base64.b64decode(password_node.getAttribute('saltValue'))
    passwordHashAlgorithm = password_node.getAttribute('hashAlgorithm')
    passwordKeyBits = int(password_node.getAttribute('keyBits'))
    info = {
        'keyDataSalt': keyDataSalt,
        'keyDataHashAlgorithm': keyDataHashAlgorithm,
        'keyDataBlockSize': keyDataBlockSize,
        'encryptedHmacKey': encryptedHmacKey,
        'encryptedHmacValue': encryptedHmacValue,
        'encryptedVerifierHashInput': encryptedVerifierHashInput,
        'encryptedVerifierHashValue': encryptedVerifierHashValue,
        'encryptedKeyValue': encryptedKeyValue,
        'spinValue': spinValue,
        'passwordSalt': passwordSalt,
        'passwordHashAlgorithm': passwordHashAlgorithm,
        'passwordKeyBits': passwordKeyBits,
    }
    return info


def _parseinfo(ole):
    versionMajor, versionMinor = unpack('<HH', ole.read(4))
    if versionMajor == 4 and versionMinor == 4:  # Agile
        return 'agile', _parseinfo_agile(ole)
    elif versionMajor in [2, 3, 4] and versionMinor == 2:  # Standard
        return 'standard', _parseinfo_standard(ole)
    elif versionMajor in [3, 4] and versionMinor == 3:  # Extensible
        raise Exception("Unsupported EncryptionInfo version (Extensible Encryption)")


class OOXMLFile(base.BaseOfficeFile):
    def __init__(self, file):
        self.format = "ooxml"
        file.seek(0)  # TODO: Investigate the effect (required for olefile.isOleFile)
        # olefile cannot process non password protected ooxml files.
        # TODO: this code is duplicate of OfficeFile(). Merge?
        if olefile.isOleFile(file):
            ole = olefile.OleFileIO(file)
            self.file = ole
            with self.file.openstream('EncryptionInfo') as stream:
                self.type, self.info = _parseinfo(stream)
            logger.debug("OOXMLFile.type: {}".format(self.type))
            self.secret_key = None
            if self.type == 'agile':
                # TODO: Support aliases?
                self.keyTypes = ('password', 'private_key', 'secret_key')
            elif self.type == 'standard':
                self.keyTypes = ('password', 'secret_key')
            elif self.type == 'extensible':
                pass
        elif zipfile.is_zipfile(file):
            self.file = file
            self.type, self.info = None, None
            self.secret_key = None
        else:
            raise Exception("Unsupported file format")

    def load_key(self, password=None, private_key=None, secret_key=None, verify_password=False):
        if password:
            if self.type == 'agile':
                self.secret_key = ECMA376Agile.makekey_from_password(
                    password,
                    self.info['passwordSalt'],
                    self.info['passwordHashAlgorithm'],
                    self.info['encryptedKeyValue'],
                    self.info['spinValue'],
                    self.info['passwordKeyBits']
                )
                if verify_password:
                    verified = ECMA376Agile.verify_password(
                        password,
                        self.info['passwordSalt'],
                        self.info['passwordHashAlgorithm'],
                        self.info['encryptedVerifierHashInput'],
                        self.info['encryptedVerifierHashValue'],
                        self.info['spinValue'],
                        self.info['passwordKeyBits']
                    )
                    if not verified:
                        raise Exception("Key verification failed")
            elif self.type == 'standard':
                self.secret_key = ECMA376Standard.makekey_from_password(
                    password,
                    self.info['header']['algId'],
                    self.info['header']['algIdHash'],
                    self.info['header']['providerType'],
                    self.info['header']['keySize'],
                    self.info['verifier']['saltSize'],
                    self.info['verifier']['salt']
                )
                if verify_password:
                    verified = ECMA376Standard.verifykey(
                        self.secret_key,
                        self.info['verifier']['encryptedVerifier'],
                        self.info['verifier']['encryptedVerifierHash']
                    )
                    if not verified:
                        raise Exception("Key verification failed")
            elif self.type == 'extensible':
                pass
        elif private_key:
            if self.type == 'agile':
                self.secret_key = ECMA376Agile.makekey_from_privkey(private_key, self.info['encryptedKeyValue'])
            else:
                raise Exception("Unsupported key type for the encryption method")
        elif secret_key:
            self.secret_key = secret_key

    def decrypt(self, ofile, verify_integrity=False):
        if self.type == 'agile':
            with self.file.openstream('EncryptedPackage') as stream:
                if verify_integrity:
                    verified = ECMA376Agile.verify_integrity(
                        self.secret_key,
                        self.info['keyDataSalt'],
                        self.info['keyDataHashAlgorithm'],
                        self.info['keyDataBlockSize'],
                        self.info['encryptedHmacKey'],
                        self.info['encryptedHmacValue'],
                        stream,
                    )
                    if not verified:
                        raise Exception('Payload integrity verification failed')

                obuf = ECMA376Agile.decrypt(
                    self.secret_key, self.info['keyDataSalt'],
                    self.info['keyDataHashAlgorithm'],
                    stream
                )
            ofile.write(obuf)
        elif self.type == 'standard':
            with self.file.openstream('EncryptedPackage') as stream:
                obuf = ECMA376Standard.decrypt(self.secret_key, stream)
            ofile.write(obuf)

        # If the file is successfully decrypted, there must be a valid OOXML file, i.e. a valid zip file
        if not zipfile.is_zipfile(io.BytesIO(obuf)):
            raise Exception("The file could not be decrypted with this password")

    def is_encrypted(self):
        # Heuristic
        if isinstance(self.file, olefile.OleFileIO):
            return True
        else:
            return False



import functools, io, logging
from hashlib import sha1
from struct import pack

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _makekey(password, salt, keyLength, block, algIdHash=0x00008004):
    r'''
    Return a intermediate key.
    '''
    # https://msdn.microsoft.com/en-us/library/dd920677(v=office.12).aspx
    password = password.encode("UTF-16LE")
    h0 = sha1(salt + password).digest()
    blockbytes = pack("<I", block)
    hfinal = sha1(h0 + blockbytes).digest()
    if keyLength == 40:
        key = hfinal[:5] + b"\x00" * 11
    else:
        key = hfinal[:keyLength // 8]
    return key


class DocumentRC4CryptoAPI:
    def __init__(self):
        pass

    @staticmethod
    def verifypw(password, salt, keySize, encryptedVerifier, encryptedVerifierHash, algId=0x00006801, block=0):
        r'''
        Return True if the given password is valid.
        '''
        # TODO: For consistency with others, rename method to verify_password or the like
        # https://msdn.microsoft.com/en-us/library/dd953617(v=office.12).aspx
        key = _makekey(password, salt, keySize, block)
        cipher = Cipher(algorithms.ARC4(key), mode=None, backend=default_backend())
        decryptor = cipher.decryptor()
        verifier = decryptor.update(encryptedVerifier)
        verfiferHash = decryptor.update(encryptedVerifierHash)
        hash = sha1(verifier).digest()
        logging.debug([verfiferHash, hash])
        return hash == verfiferHash

    @staticmethod
    def decrypt(password, salt, keySize, ibuf, blocksize=0x200, block=0):
        r'''
        Return decrypted data.
        '''
        obuf = io.BytesIO()

        key = _makekey(password, salt, keySize, block)

        for c, buf in enumerate(iter(functools.partial(ibuf.read, blocksize), b'')):
            cipher = Cipher(algorithms.ARC4(key), mode=None, backend=default_backend())
            decryptor = cipher.decryptor()

            dec = decryptor.update(buf) + decryptor.finalize()
            obuf.write(dec)

            # From wvDecrypt:
            # at this stage we need to rekey the rc4 algorithm
            # Dieter Spaar <spaar@mirider.augusta.de> figured out
            # this rekeying, big kudos to him
            block += 1
            key = _makekey(password, salt, keySize, block)

        obuf.seek(0)
        return obuf



import functools
import hmac
import io
import logging
from hashlib import sha1, sha512
from struct import pack, unpack

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _get_hash_func(algorithm):
    if algorithm == 'SHA512':
        return sha512
    else:
        return sha1


def _decrypt_aes_cbc(data, key, iv):
    aes = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
    decryptor = aes.decryptor()
    decrypted = decryptor.update(data) + decryptor.finalize()
    return decrypted


class ECMA376Agile:
    def __init__(self):
        pass

    @staticmethod
    def _derive_iterated_hash_from_password(password, saltValue, hashAlgorithm, spinValue):
        r'''
        Do a partial password-based hash derivation.
        Note the block key is not taken into consideration in this function.
        '''
        # TODO: This function is quite expensive and it should only be called once.
        # We need to save the result for later use.
        # This is not covered by the specification, but MS Word does so.

        hashCalc = _get_hash_func(hashAlgorithm)

        # NOTE: Initial round sha512(salt + password)
        h = hashCalc(saltValue + password.encode("UTF-16LE"))

        # NOTE: Iteration of 0 -> spincount-1; hash = sha512(iterator + hash)
        for i in range(0, spinValue, 1):
            h = hashCalc(pack("<I", i) + h.digest())

        return h

    @staticmethod
    def _derive_encryption_key(h, blockKey, hashAlgorithm, keyBits):
        r'''
        Finish the password-based key derivation by hashing last hash + blockKey.
        '''
        hashCalc = _get_hash_func(hashAlgorithm)
        h_final = hashCalc(h + blockKey)

        # NOTE: Needed to truncate encryption key to bitsize
        encryption_key = h_final.digest()[:keyBits // 8]

        return encryption_key

    @staticmethod
    def decrypt(key, keyDataSalt, hashAlgorithm, ibuf):
        r'''
        Return decrypted data.

            >>> key = b'@ f\t\xd9\xfa\xad\xf2K\x07j\xeb\xf2\xc45\xb7B\x92\xc8\xb8\xa7\xaa\x81\xbcg\x9b\xe8\x97\x11\xb0*\xc2'
            >>> keyDataSalt = b'\x8f\xc7x"+P\x8d\xdcL\xe6\x8c\xdd\x15<\x16\xb4'
            >>> hashAlgorithm = 'SHA512'
        '''
        SEGMENT_LENGTH = 4096
        hashCalc = _get_hash_func(hashAlgorithm)

        obuf = io.BytesIO()
        totalSize = unpack('<I', ibuf.read(4))[0]
        logger.debug("totalSize: {}".format(totalSize))
        remaining = totalSize
        ibuf.seek(8)
        for i, buf in enumerate(iter(functools.partial(ibuf.read, SEGMENT_LENGTH), b'')):
            saltWithBlockKey = keyDataSalt + pack('<I', i)
            iv = hashCalc(saltWithBlockKey).digest()
            iv = iv[:16]
            aes = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
            decryptor = aes.decryptor()
            dec = decryptor.update(buf) + decryptor.finalize()
            if remaining < len(buf):
                dec = dec[:remaining]
            obuf.write(dec)
            remaining -= len(buf)
        return obuf.getvalue()  # return obuf.getbuffer()

    @staticmethod
    def verify_password(password, saltValue, hashAlgorithm, encryptedVerifierHashInput, encryptedVerifierHashValue, spinValue, keyBits):
        r'''
        Return True if the given password is valid.

            >>> password = 'Password1234_'
            >>> saltValue = b'\xcb\xca\x1c\x99\x93C\xfb\xad\x92\x07V4\x15\x004\xb0'
            >>> hashAlgorithm = 'SHA512'
            >>> encryptedVerifierHashInput = b'9\xee\xa5N&\xe5\x14y\x8c(K\xc7qM8\xac'
            >>> encryptedVerifierHashValue = b'\x147mm\x81s4\xe6\xb0\xffO\xd8"\x1a|g\x8e]\x8axN\x8f\x99\x9fL\x18\x890\xc3jK)\xc5\xb33`' + \
            ... b'[\\\xd4\x03\xb0P\x03\xad\xcf\x18\xcc\xa8\xcb\xab\x8d\xeb\xe3s\xc6V\x04\xa0\xbe\xcf\xae\\\n\xd0'
            >>> spinValue = 100000
            >>> keyBits = 256
            >>> ECMA376Agile.verify_password(password, saltValue, hashAlgorithm, encryptedVerifierHashInput, encryptedVerifierHashValue, spinValue, keyBits)
            True
        '''
        # NOTE: See https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-offcrypto/a57cb947-554f-4e5e-b150-3f2978225e92

        block1 = bytearray([0xfe, 0xa7, 0xd2, 0x76, 0x3b, 0x4b, 0x9e, 0x79])
        block2 = bytearray([0xd7, 0xaa, 0x0f, 0x6d, 0x30, 0x61, 0x34, 0x4e])

        h = ECMA376Agile._derive_iterated_hash_from_password(password, saltValue, hashAlgorithm, spinValue)

        key1 = ECMA376Agile._derive_encryption_key(h.digest(), block1, hashAlgorithm, keyBits)
        key2 = ECMA376Agile._derive_encryption_key(h.digest(), block2, hashAlgorithm, keyBits)

        hash_input = _decrypt_aes_cbc(encryptedVerifierHashInput, key1, saltValue)
        hashCalc = _get_hash_func(hashAlgorithm)
        acutal_hash = hashCalc(hash_input)
        acutal_hash = acutal_hash.digest()

        expected_hash = _decrypt_aes_cbc(encryptedVerifierHashValue, key2, saltValue)

        return acutal_hash == expected_hash

    @staticmethod
    def verify_integrity(secretKey, keyDataSalt, keyDataHashAlgorithm, keyDataBlockSize, encryptedHmacKey, encryptedHmacValue, stream):
        r'''
        Return True if the HMAC of the data payload is valid.
        '''
        # NOTE: See https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-offcrypto/63d9c262-82b9-4fa3-a06d-d087b93e3b00

        block4 = bytearray([0x5f, 0xb2, 0xad, 0x01, 0x0c, 0xb9, 0xe1, 0xf6])
        block5 = bytearray([0xa0, 0x67, 0x7f, 0x02, 0xb2, 0x2c, 0x84, 0x33])

        hashCalc = _get_hash_func(keyDataHashAlgorithm)

        iv1 = hashCalc(keyDataSalt + block4).digest()
        iv1 = iv1[: keyDataBlockSize]
        iv2 = hashCalc(keyDataSalt + block5).digest()
        iv2 = iv2[: keyDataBlockSize]

        hmacKey = _decrypt_aes_cbc(encryptedHmacKey, secretKey, iv1)
        hmacValue = _decrypt_aes_cbc(encryptedHmacValue, secretKey, iv2)

        msg_hmac = hmac.new(hmacKey, stream.read(), hashCalc)
        actualHmac = msg_hmac.digest()
        stream.seek(0)

        return hmacValue == actualHmac

    @staticmethod
    def makekey_from_privkey(privkey, encryptedKeyValue):
        privkey = serialization.load_pem_private_key(privkey.read(), password=None, backend=default_backend())
        skey = privkey.decrypt(encryptedKeyValue, padding.PKCS1v15())
        return skey

    @staticmethod
    def makekey_from_password(password, saltValue, hashAlgorithm, encryptedKeyValue, spinValue, keyBits):
        r'''
        Generate intermediate key from given password.

            >>> password = 'Password1234_'
            >>> saltValue = b'Lr]E\xdca\x0f\x93\x94\x12\xa0M\xa7\x91\x04f'
            >>> hashAlgorithm = 'SHA512'
            >>> encryptedKeyValue = b"\xa1l\xd5\x16Zz\xb9\xd2q\x11>\xd3\x86\xa7\x8c\xf4\x96\x92\xe8\xe5'\xb0\xc5\xfc\x00U\xed\x08\x0b|\xb9K"
            >>> spinValue = 100000
            >>> keyBits = 256
            >>> expected = b'@ f\t\xd9\xfa\xad\xf2K\x07j\xeb\xf2\xc45\xb7B\x92\xc8\xb8\xa7\xaa\x81\xbcg\x9b\xe8\x97\x11\xb0*\xc2'
            >>> ECMA376Agile.makekey_from_password(password, saltValue, hashAlgorithm, encryptedKeyValue, spinValue, keyBits) == expected
            True
        '''
        block3 = bytearray([0x14, 0x6e, 0x0b, 0xe7, 0xab, 0xac, 0xd0, 0xd6])

        h = ECMA376Agile._derive_iterated_hash_from_password(password, saltValue, hashAlgorithm, spinValue)
        encryption_key = ECMA376Agile._derive_encryption_key(h.digest(), block3, hashAlgorithm, keyBits)

        skey = _decrypt_aes_cbc(encryptedKeyValue, encryption_key, saltValue)

        return skey



import logging
from struct import unpack

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# https://msdn.microsoft.com/en-us/library/dd926359(v=office.12).aspx
def _parse_encryptionheader(blob):
    flags, = unpack('<I', blob.read(4))
    # if mode == 'strict': compare values with spec.
    sizeExtra, = unpack('<I', blob.read(4))
    algId, = unpack('<I', blob.read(4))
    algIdHash, = unpack('<I', blob.read(4))
    keySize, = unpack('<I', blob.read(4))
    providerType, = unpack('<I', blob.read(4))
    reserved1, = unpack('<I', blob.read(4))
    reserved2, = unpack('<I', blob.read(4))
    cspName = blob.read().decode('utf-16le')
    header = {
        'flags': flags,
        'sizeExtra': sizeExtra,
        'algId': algId,
        'algIdHash': algIdHash,
        'keySize': keySize,
        'providerType': providerType,
        'reserved1': reserved1,
        'reserved2': reserved2,
        'cspName': cspName,
    }
    return header


# https://msdn.microsoft.com/en-us/library/dd910568(v=office.12).aspx
def _parse_encryptionverifier(blob, algorithm):
    saltSize, = unpack('<I', blob.read(4))
    salt = blob.read(16)
    encryptedVerifier = blob.read(16)
    verifierHashSize, = unpack('<I', blob.read(4))
    if algorithm == 'RC4':
        encryptedVerifierHash = blob.read(20)
    elif algorithm == 'AES':
        encryptedVerifierHash = blob.read(32)
    verifier = {
        'saltSize': saltSize,
        'salt': salt,
        'encryptedVerifier': encryptedVerifier,
        'verifierHashSize': verifierHashSize,
        'encryptedVerifierHash': encryptedVerifierHash,
    }
    return verifier



import functools, io, logging
from hashlib import md5
from struct import pack

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _makekey(password, salt, block):
    r'''
    Return a intermediate key.

        >>> password = 'password1'
        >>> salt = b'\xe8w,\x1d\x91\xc5j7\x96Ga\xb2\x80\x182\x17'
        >>> block = 0
        >>> expected = b' \xbf2\xdd\xf5@\x85\x8cQ7D\xaf\x0f$\xe0<'
        >>> _makekey(password, salt, block) == expected
        True
    '''
    # https://msdn.microsoft.com/en-us/library/dd920360(v=office.12).aspx
    password = password.encode("UTF-16LE")
    h0 = md5(password).digest()
    truncatedHash = h0[:5]
    intermediateBuffer = (truncatedHash + salt) * 16
    h1 = md5(intermediateBuffer).digest()
    truncatedHash = h1[:5]
    blockbytes = pack("<I", block)
    hfinal = md5(truncatedHash + blockbytes).digest()
    key = hfinal[:128 // 8]
    return key


class DocumentRC4:
    def __init__(self):
        pass

    @staticmethod
    def verifypw(password, salt, encryptedVerifier, encryptedVerifierHash):
        r'''
        Return True if the given password is valid.

            >>> password = 'password1'
            >>> salt = b'\xe8w,\x1d\x91\xc5j7\x96Ga\xb2\x80\x182\x17'
            >>> encryptedVerifier = b'\xc9\xe9\x97\xd4T\x97=1\x0b\xb1\xbap\x14&\x83~'
            >>> encryptedVerifierHash = b'\xb1\xde\x17\x8f\x07\xe9\x89\xc4M\xae^L\xf9j\xc4\x07'
            >>> DocumentRC4.verifypw(password, salt, encryptedVerifier, encryptedVerifierHash)
            True
        '''
        # https://msdn.microsoft.com/en-us/library/dd952648(v=office.12).aspx
        block = 0
        key = _makekey(password, salt, block)
        cipher = Cipher(algorithms.ARC4(key), mode=None, backend=default_backend())
        decryptor = cipher.decryptor()
        verifier = decryptor.update(encryptedVerifier)
        verfiferHash = decryptor.update(encryptedVerifierHash)
        hash = md5(verifier).digest()
        logging.debug([verfiferHash, hash])
        return hash == verfiferHash

    @staticmethod
    def decrypt(password, salt, ibuf, blocksize=0x200):
        r'''
        Return decrypted data.
        '''
        obuf = io.BytesIO()

        block = 0
        key = _makekey(password, salt, block)

        for c, buf in enumerate(iter(functools.partial(ibuf.read, blocksize), b'')):
            cipher = Cipher(algorithms.ARC4(key), mode=None, backend=default_backend())
            decryptor = cipher.decryptor()

            dec = decryptor.update(buf) + decryptor.finalize()
            obuf.write(dec)

            # From wvDecrypt:
            # at this stage we need to rekey the rc4 algorithm
            # Dieter Spaar <spaar@mirider.augusta.de> figured out
            # this rekeying, big kudos to him
            block += 1
            key = _makekey(password, salt, block)

        obuf.seek(0)
        return obuf



from setuptools import setup
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

# https://packaging.python.org/guides/single-sourcing-package-version/
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='msoffcrypto-tool',
    version=find_version("msoffcrypto", "__init__.py"),
    description='A Python tool and library for decrypting MS Office files with passwords or other keys',
    long_description=open("README.md", "r").read(),
    long_description_content_type='text/markdown',
    url='https://github.com/nolze/msoffcrypto-tool',
    author='nolze',
    author_email='nolze@archlinux.us',
    license='MIT',
    keywords='',
    packages=[
        "msoffcrypto",
        "msoffcrypto.format",
        "msoffcrypto.method",
    ],
    install_requires=[
        'olefile >= 0.45',
        'cryptography >= 2.3',
    ],
    tests_require=[
        'nose >= 1.3.7',
        'coverage >= 4.5.1', 
    ],
    classifiers=[
    ],
    entry_points={
        'console_scripts': [
            'msoffcrypto-tool = msoffcrypto.__main__:main',
        ],
    },
)
